# -*- coding: utf-8 -*-
"""
Module et_micc.static_vars
=======================
A decorator for adding static variables to a function.
see https://stackoverflow.com/questions/279561/what-is-the-python-equivalent-of-static-variables-inside-a-function
"""

def static_vars(**kwargs):
    """Add static variables to a method.
    
    Use as:
    
    .. code-block:: python
       @static_vars(counter=0)
       def foo():
           foo.counter += 1
    
    """
    def decorate(func):
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func
    return decorate

#eof


"""
Module f2py
===========
Wrapper for Numpy.f2py
"""

from pathlib import Path

import numpy.f2py

def build_f2py(module_name,args=[]):
    """
    :param Path path: to f90 source
    """
    src_file = module_name + '.f90'
    
    path_to_src_file = Path(src_file).resolve()
    if not path_to_src_file.exists():
        raise FileNotFoundError(str(path_to_src_file))
    
    f2py_args = ['--build-dir','_f2py_build']
    f2py_args .extend(args)
    
    with open(str(path_to_src_file.name)) as f:
        fsource = f.read() 
    returncode = numpy.f2py.compile(fsource, extension='.f90', modulename=module_name, extra_args=f2py_args, verbose=True)
    
    return returncode


# -*- coding: utf-8 -*-
"""Command line interface build (no sub-commands)."""


import os,sys
from types import SimpleNamespace
from pathlib import Path

import click

import et_micc_tools.commands as cmds
# import et_micc_tools.utils
import et_micc_tools.logging_tools
import et_micc_tools.expand

__template_help  =  "Ordered list of Cookiecutter templates, or a single Cookiecutter template."

__micc_file_help = ("The path to the *micc-file* with the parameter values used in the cookiecutter"
                    "templates. When a new project is created, "
                    "in the cookiecutter templates (default = ``micc.json``). "
                    "*Micc* does not use the standard ``cookiecutter.json`` file to provide the "
                    "template parameters, but uses a *micc-file*, usually named ``micc.json``, to "
                    "generate a ``cookiecutter.json`` file. Unlike the ``cookiecutter.json`` file, "
                    "the ``micc.json`` file can contain default values for the template parameters. "
                    "You will be prompted to provide a value for all parameters without default value. "
                    "*Micc* looks for the *micc-file* in the template directory **only** "
                    "(as specified with the ``--template`` option)."
                   )


@click.group()
@click.option('-v', '--verbosity', count=True
             , help="The verbosity of the program output."
             , default=1
             )
@click.option('-p', '--project-path'
             , help="The path to the project directory. "
                    "The default is the current working directory."
             , default='.'
             , type=Path
             )
@click.option('--clear-log'
             , help="If specified clears the project's ``et_micc_tools.log`` file."
             , default=False, is_flag=True
             )
@click.pass_context
def main(ctx, verbosity, project_path, clear_log):
    """Micc command line interface.
    
    All commands that change the state of the project produce some output that
    is send to the console (taking verbosity into account). It is also sent to
    a logfile ``et_micc_tools.log`` in the project directory. All output is always appended
    to the logfile. If you think the file has gotten too big, or you are no more
    interested in the history of your project, you can specify the ``--clear-log``
    flag to clear the logfile before any command is executed. In this way the
    command you execute is logged to an empty logfile.
    
    See below for (sub)commands.
    """
    if clear_log:
        os.remove(project_path / 'micc.log')
        
    ctx.obj = SimpleNamespace( verbosity=verbosity
                             , project_path=project_path.resolve()
                             , clear_log=clear_log
                             , template_parameters={}
                             )

#     if et_micc_tools.utils.is_conda_python():
#         click.echo( click.style("==========================================================\n"
#                                 "WARNING: You are running in a conda Python environment.\n"
#                                 "         Note that poetry does not play well with conda.\n",   fg='yellow')
#                   + click.style("         Especially, do NOT use:\n"
#                                 "         >  poetry install\n",                                 fg='bright_red')
#                   + click.style("==========================================================\n", fg='yellow')
#                   )
    template_parameters_json = project_path / 'micc.json'
    if template_parameters_json.exists():
        ctx.obj.template_parameters.update(
            et_micc_tools.expand.get_template_parameters(template_parameters_json)
        )
    else:
        ctx.obj.template_parameters.update(
            et_micc_tools.expand.get_template_parameters(
                et_micc_tools.expand.get_preferences(Path('.'))
            )
        )
    

def _check_cxx_flags(cxx_flags,cli_option):
    """
    :param str cxx_flags: C++ compiler flags
    :param str cli_option: typically '--cxx-flags', or '--cxx-flags-all'.
    :raises: RunTimeError if cxx_flags starts or ends with a '"' but not both.
    """
    if cxx_flags.startswith('"') and cxx_flags.endswith('"'):
        # compile options appear between quotes
        pass
    elif not cxx_flags.startswith('"') and not cxx_flags.endswith('"'):
        # a singlecompile option must still be surrounded with quotes. 
        cxx_flags = f'"{cxx_flags}"'
    else:
        raise RuntimeError(f"{cli_option}: unmatched quotes: {cxx_flags}")
    return cxx_flags


def _check_load_save(filename,loadorsave):
    """
    :param str filename: possibly empty string.
    :param str loadorsave: 'load'|'save'.
    :raises: RunTimeError if filename is actually a file path.
    """
    if filename:
        if os.sep in filename:
            raise RuntimeError(f"--{loadorsave} {filename}: only filename allowed, not path.")
        if not filename.endswith('.json'):
            filename += '.json'
    return filename


@main.command()
@click.option('-m','--module'
             , help="Build only this module. The module kind prefix (``cpp_`` "
                    "for C++ modules, ``f2py_`` for Fortran modules) may be omitted."
             , default=''
             )
@click.option('-b','--build-type'
             , help="build type: For f2py modules, either RELEASE or DEBUG, where the latter"
                    "toggles the --debug, --noopt, and --noarch, and ignores all other "
                    "f2py compile flags. For cpp modules any of the standard CMake build types: "
                    "DEBUG, MINSIZEREL, RELEASE, RELWITHDEBINFO."
             , default='RELEASE'
             )
# F2py specific options
@click.option('--f90exec'
             , help="F2py: Specify the path to F90 compiler."
             , default=''
             )
@click.option('--f90flags'
             , help="F2py: Specify F90 compiler flags."
             , default='-O3'
             )
@click.option('--opt'
             , help="F2py: Specify optimization flags."
             , default='' 
             )
@click.option('--arch'
             , help="F2py: Specify architecture specific optimization flags."
             , default='' 
             )
@click.option('--debug'
             , help="F2py: Compile with debugging information."
             , default=False, is_flag=True
             )
@click.option('--noopt'
             , help="F2py: Compile without optimization."
             , default=False, is_flag=True
             )
@click.option('--noarch'
             , help="F2py: Compile without architecture specific optimization."
             , default=False, is_flag=True
             )
# Cpp specific options
@click.option('--cxx-compiler'
             , help="CMake: specify the C++ compiler (sets CMAKE_CXX_COMPILER)."
             , default=''
             )
@click.option('--cxx-flags'
             , help="CMake: set CMAKE_CXX_FLAGS_<built_type> to <cxx_flags>."
             , default=''
             )
@click.option('--cxx-flags-all'
             , help="CMake: set CMAKE_CXX_FLAGS_<built_type> to <cxx_flags>."
             , default=''
             )
# Other options
@click.option('--clean'
             , help="Perform a clean build."
             , default=False, is_flag=True
             )
@click.option('--load'
             , help="Load the build options from a (.json) file in the module directory. "
                    "All other compile options are ignored."
             , default=''
             )
@click.option('--save'
             , help="Save the build options to a (.json) file in the module directory."
             , default=''
             )
@click.option('-s', '--soft-link'
             , help="Create a soft link rather than a copy of the binary extension module."
             , default=False, is_flag=True
             )
@click.pass_context
def build( ctx, module
         , build_type
# F2py specific options
         , f90exec
         , f90flags, opt, arch
         , debug, noopt, noarch
# Cpp specific options
         , cxx_compiler
         , cxx_flags, cxx_flags_all
# Other options
         , clean
         , soft_link
         , load, save
         ):
    """Build binary extension libraries (f2py and cpp modules)."""
    if save:
        if os.sep in save:
            raise RuntimeError(f"--save {save}: only filename allowed, not path.")
        if not save.endswith('.json'):
            save += '.json'
    if load:
        if os.sep in load:
            raise RuntimeError(f"--load {load}: only filename allowed, not path.")
        if not load.endswith('.json'):
            load += '.json'
    
    with et_micc_tools.logging_tools.logtime(ctx.obj):            
        build_options=SimpleNamespace( build_type = build_type.upper() )
        build_options.clean = clean
        build_options.soft_link = soft_link
        build_options.save = _check_load_save(save, "save")
        build_options.load = _check_load_save(load, "load")
        if not load:
            if build_type=='DEBUG':
                f2py = {'--debug' :None
                       ,'--noopt' :None
                       ,'--noarch':None
                       }
            else:
                f2py = {}
                if f90exec:
                    f2py['--f90exec'] = f90exec
                if f90flags:
                    f2py['--f90flags'] = f90flags
                if opt:
                    f2py['--opt'] = opt
                if arch:
                    f2py['--arch'] = arch
                if noopt:
                    f2py['--noopt'] = None
                if noarch:
                    f2py['--noarch'] = None
                if debug:
                    f2py['--debug'] = None
            build_options.f2py = f2py
    
            cmake = {}
            cmake['CMAKE_BUILD_TYPE'] = build_type
            if cxx_compiler:
                path_to_cxx_compiler = Path(cxx_compiler).resolve()
                if not path_to_cxx_compiler.exists():
                    raise FileNotFoundError(f"C++ compiler {path_to_cxx_compiler} not found.")
                cmake['CMAKE_CXX_COMPILER'] = str(path_to_cxx_compiler)
            if cxx_flags:
                cmake[f"CMAKE_CXX_FLAGS_{build_type}"] = _check_cxx_flags(cxx_flags,"--cxx-flags")
            if cxx_flags_all:
                cmake["CMAKE_CXX_FLAGS"] = _check_cxx_flags(cxx_flags_all,"--cxx-flags-all")
            build_options.cmake = cmake
            
        ctx.obj.build_options = build_options
        
        rc = cmds.micc_build( module_to_build=module
                            , global_options=ctx.obj
                            )
    if rc:
        ctx.exit(rc)
      

if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover
#eodf



# -*- coding: utf-8 -*-

"""
Module et_micc_tools_tools.commands 
=================================================================

A module

"""
import os
import json
import shutil


import et_micc_tools.utils
import et_micc_tools.logging_tools
from et_micc_tools.f2py import build_f2py
from et_micc_tools.tomlfile import TomlFile


def micc_build( module_to_build, global_options ):
    """
    Build binary extensions, i.e. f2py modules and cpp modules.
    
    :param str module_to_build: name of the only module to build (the prefix 
        ``cpp_`` or ``f2py_`` may be omitted). If not provided, all binrary
        extensions are built.
    :param types.SimpleNamespace global_options: namespace object with
        options accepted by (almost) all et_micc_tools commands. Relevant attributes are 
        
        * **verbosity**
        * **project_path**: Path to the project on which the command operates.
        * **build_options**: all build options.
    """
    project_path = global_options.project_path

    et_micc_tools.utils.is_project_directory(project_path, raise_if=False)
    et_micc_tools.utils.is_package_project  (project_path, raise_if=False)
    et_micc_tools.utils.is_module_project   (project_path, raise_if=True )
    
    build_options = global_options.build_options
    
    package_path = project_path / et_micc_tools.utils.convert_to_valid_module_name(project_path.name)
        
    # get extension for binary extensions (depends on OS and python version)
    extension_suffix = et_micc_tools.utils.get_extension_suffix()
    
    dirs = os.listdir(package_path)
    for d in dirs:
        if (     (package_path / d).is_dir() 
             and (d.startswith("f2py_") or d.startswith("cpp_"))
           ):
            if module_to_build and not d.endswith(module_to_build): # build only this module
                continue

            build_log_file = project_path / f"et_micc_tools-build-{d}.log"
            build_logger = et_micc_tools.logging_tools.create_logger( build_log_file, filemode='w' )
 
            module_type,module_name = d.split('_',1)
            
            with et_micc_tools.logging_tools.log(build_logger.info,f"Building {module_type} module {module_name}"):
                cextension = module_name + extension_suffix
                destination = (package_path / cextension).resolve()
                if build_options.clean:
                    os.remove(str(destination))
                module_dir = package_path / d 
                if build_options.save:
                    with open(str(module_dir / build_options.save),'w') as f:
                        json.dump(build_options.f2py,f)
                
                if module_type=='f2py':
                    if build_options.load:
                        with open(str(module_dir / build_options.save),'r') as f:
                            build_options.f2py = json.load(f)
                    build_dir = module_dir
                    f2py_args = []
                    for arg,val in build_options.f2py.items():
                        if val is None:
                            # this is a flag
                            f2py_args.append(arg)
                        else:
                            f2py_args.append(f"{arg}=\"{val}\"")

                    if build_options.save:
                        with open(str(module_dir / build_options.save),'w') as f:
                            json.dump(build_options.f2py,f)
                    with et_micc_tools.utils.in_directory(module_dir):
                        if build_options.clean:
                            build_logger.info(f"--clean: removing {d}/_f2py_build")
                            shutil.rmtree('_f2py_build') 
                        returncode = build_f2py(module_name, args=f2py_args)
                    
                elif module_type=='cpp':
                    if build_options.load:
                        with open(str(module_dir / build_options.save),'r') as f:
                            build_options.cmake = json.load(f)
                    build_dir = module_dir  / '_cmake_build'
                    if build_options.clean:
                        build_logger.info(f"--clean: removing {d}/_cmake_build")
                        shutil.rmtree(build_dir) 
                    build_dir.mkdir(parents=True, exist_ok=True)
                    with et_micc_tools.utils.in_directory(build_dir):
                        cmake_cmd = ['poetry','run','cmake','-D',f"pybind11_DIR={et_micc_tools.utils.path_to_cmake_tools()}"]
                        for key,val in build_options.cmake.items():
                            cmake_cmd.extend(['-D',f"{key}={val}"])
                        cmake_cmd.append('..')
                        cmds = [ cmake_cmd
                               , ['make']
                               ]
                        # WARNING: for these commands to work in eclipse, eclipse must have
                        # started from the shell with the appropriate environment activated.
                        # Otherwise subprocess starts out with the wrong environment. It 
                        # may not pick the right Python version, and may not find pybind11.
                        returncode = et_micc_tools.utils.execute(cmds, build_logger.debug, stop_on_error=True, env=os.environ.copy())
                else:
                    raise RuntimeError(f"Unknown module_type: {module_type}")

                if returncode:
                    return returncode 
                
                built = build_dir / cextension
                destination = (package_path / cextension).resolve()
                if build_options.soft_link:
                    cmds = ['ln', '-sf', str(built), str(destination)]
                    returncode = et_micc_tools.utils.execute(cmds, build_logger.debug, stop_on_error=True, env=os.environ.copy())
                    if returncode:
                        return returncode 
                else:
                    if destination.exists():
                        build_logger.debug(f">>> os.remove({destination})\n")
                        destination.unlink()
                    build_logger.debug(f">>> shutil.copyfile( '{built}', '{destination}' )\n")
                    shutil.copyfile(built, destination)
                
                    # Remove the build directory to avoid that it will be included in the wheel
                    # (we cannot do this if build_options.soft_link is True 
                    if module_type=='f2py':
                        build_dir = module_dir / '_f2py_build'
                    shutil.rmtree(build_dir) 
                
            build_logger.info(f"Built: {destination}\n"
                              f"Check {build_log_file} for details."
                             )
    
    
    return 0


#eof


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['et_micc_tools']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0',
 'cmake>=3.15.3,<4.0.0',
 'numpy>=1.17.4,<2.0.0',
 'pybind11>=2.4.3,<3.0.0',
 'tomlkit>=0.5.5,<0.6.0']

entry_points = \
{'console_scripts': ['build = et_micc_tools:cli_build']}

setup_kwargs = {
    'name': 'et-micc-tools',
    'version': '0.0.4',
    'description': '<Enter a one-sentence description of this project here.>',
    'long_description': '=============\net-micc-tools\n=============\n\n\n\n<Enter a one-sentence description of this project here.>\n\n\n* Free software: MIT license\n* Documentation: https://et-micc-tools.readthedocs.io.\n\n\nFeatures\n--------\n\n* TODO\n',
    'author': 'Engelbert Tijskens',
    'author_email': 'engelbert.tijskens@uantwerpen.be',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/etijskens/et-micc-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""
Module et_micc_tools.utils
=================

Utility functions for et_micc_tools.py.
"""
import os, sys, subprocess, logging, sysconfig, copy
from contextlib import contextmanager
# import toml
from et_micc_tools.tomlfile import TomlFile

from pathlib import Path
from datetime import datetime
import et_micc_tools.logging_tools


def get_extension_suffix():
    return sysconfig.get_config_var('EXT_SUFFIX')


def path_to_cmake_tools():
    """Return the path to the folder with the CMake tools."""
    
    p = (Path(__file__) / '..' / 'cmake_tools').resolve()
    return str(p)


@contextmanager
def in_directory(path):
    """Context manager for changing the current working directory while the body of the
    context manager executes.
    """
    previous_dir = os.getcwd()
    os.chdir(str(path)) # the str method takes care of when path is a Path object
    yield os.getcwd()
    os.chdir(previous_dir)


def replace_version_in_file(filepath,current_version,new_version):
    """Replace the version string in a file.
    
    :param Path filepath: Path to the file.
    """
    if filepath.exists():
        if filepath.name == "pyproject.toml":
            fmt = 'version = "{}"'
        else:
            fmt = '__version__ = "{}"'
        old = fmt.format(current_version)
        new = fmt.format(new_version)
        replace_in_file(filepath,old,new)
        return True
    else:
        return False


def replace_in_file(filepath,old,new):
    """Replace :py:obj:`old` string with :py:obj:`new` in a file.
    
    :param Path filepath: 
    """
    with filepath.open() as f:
        content_as_string = f.read()
    content_as_string = content_as_string.replace(old,new)
    with filepath.open(mode="w") as f:
        f.write(content_as_string)


def is_project_directory(path,raise_if=None):
    """Verify that the directory :file:`path` is a project directory. 
    
    As a sufficident condition, we request that 
    
    * there is a pyproject.toml file, exposing ``['tool']['poetry']['name']``
    * that there is a python package or module with that name.
    
    :param Path path:
    :param bool raise_if: If True, raise ``RuntimeError`` if the test succeeds.
        If False, raise ``RuntimeError`` if the test fails.
        If None, do not raise.
    :returns: bool.
    :raises: RuntimeError.
    """
    if not isinstance(path, Path):
        path = Path(path)
    path_to_pyproject_toml = str(path /'pyproject.toml')
    
    try:
        toml = TomlFile(path_to_pyproject_toml)
        project_name = toml['tool']['poetry']['name']
        rv = True
    except Exception as e:
        rv = False

    if rv:
        package_name = convert_to_valid_module_name(project_name)
        if (path / package_name / '__init__.py').exists():
            # python package found
            rv = True
        elif (path / (package_name + '.py')).exists():
            rv = True
        else:
            rv = False
    
    # the stuff below could be achieve through a decorator, but it is a bit
    # complex, and we need it only three times
    if raise_if is None or raise_if!=rv:
        # must not raise
        return rv
    # must raise in all other cases
    elif raise_if==True:
        x = RuntimeError(f"Project path '{path}' is a project directory.")
        x.path = path
    elif raise_if==False:
        x = RuntimeError(f"Project path '{path}' is NOT an project directory.")
        x.path = path
    raise x


def convert_to_valid_module_name(name):
    """Convert a module name to a PEP8 compliant module name.
    
    * lowercase
    * whitespace -> underscore 
    * dash -> underscore 
    * if leading numeric character, prepend '_'
      Names like this are discouraged by https://www.python.org/dev/peps/pep-0008/#package-and-module-names,
      but so are, names starting with a numeric character.
    """
    if name[0].isnumeric():
        name = '_'+name
    valid_module_name = name.lower().replace('-', '_').replace(' ', '_')
    return valid_module_name


def py_module_exists(project_path, module_name):
    """Test if there is already a python module with name :py:obj:`module_name` 
    in the project at :file:`project_path`.
        
    :param Path project_path: project path
    :param str module_name: module name
    :returns: bool
    """
    file = project_path / convert_to_valid_module_name(project_path.name) / f'{module_name}.py'
    return file.is_file()


def py_package_exists(project_path, module_name):
    """Test if there is already a python package with name :py:obj:`module_name` 
    in the project at :file:`project_path`.
        
    :param Path project_path: project path
    :param str module_name: module name
    :returns: bool
    """
    return (project_path / convert_to_valid_module_name(project_path.name) / module_name / '__init__.py').is_file()


def f2py_module_exists(project_path, module_name):
    """Test if there is already a f2py module with name py:obj:`module_name` in this project.
        
    :param Path project_path: project path
    :param str module_name: module name
    :returns: bool
    """
    return (project_path / convert_to_valid_module_name(project_path.name) / ('f2py_' + module_name)/ f"{ module_name}.f90").is_file() 


def cpp_module_exists(project_path, module_name):
    """Test if there is already a cpp module with name py:obj:`module_name` in this project.
        
    :param Path project_path: project path
    :param str module_name: module name
    :returns: bool
    """
    return (project_path / convert_to_valid_module_name(project_path.name) / ('cpp_' + module_name)/ f"{ module_name}.cpp").is_file() 


def module_exists(project_path, module_name):
    """Test if there is already a module with name py:obj:`module_name` in this project.

    :param Path project_path: project path
    :param str module_name: module name
    :returns: bool
    """
    return (  py_module_exists(project_path, module_name)
         or  py_package_exists(project_path, module_name)
         or  cpp_module_exists(project_path, module_name)
         or f2py_module_exists(project_path, module_name)
           )

def app_exists(project_path, app_name):
    """Test if there is already an app with name ``app_name`` in this project.
    
    * :file:`<package_name>/cli_<app_name>.py`
        
    :param Path project_path: project path
    :param str app_name: app name
    :returns: bool
    """
    return (project_path / convert_to_valid_module_name(project_path.name) / f"cli_{app_name}.py").is_file()
    

def is_module_project(project_path, raise_if=None):
    """Find out if this project is a module python project.
    
    :param Path project_path: project path
    :param bool raise_if: If True, raise ``RuntimeError`` if the test succeeds.
        If False, raise ``RuntimeError`` if the test fails.
        If None, do not raise.
    :returns: bool.
    :raises: RuntimeError.
    """
    package_name = convert_to_valid_module_name(project_path.name)
    rv = (project_path / f'{package_name}.py').is_file()
    if raise_if is None or raise_if!=rv:
        return rv
    elif raise_if==True:
        x = RuntimeError(f"Directory '{project_path}' is a module project")
        x.path = project_path
    elif raise_if==False:
        x = RuntimeError(f"Directory '{project_path}' is NOT a module project")
        x.path = project_path
    raise x
    

def is_package_project(project_path, raise_if=None):
    """Find out if this project is a package python project.
    
    :param Path project_path: project path
    """    
    package_name = convert_to_valid_module_name(project_path.name)
    rv = (project_path / package_name / '__init__.py').is_file()

    if raise_if is None or raise_if!=rv:
        return rv
    elif raise_if==True:
        x = RuntimeError(f"Directory '{project_path}' is a package project")
        x.path = project_path
    elif raise_if==False:
        x = RuntimeError(f"Directory '{project_path}' is NOT a package project")
        x.path = project_path
    raise x


def execute(cmds,logfun=None,stop_on_error=True,env=None):
    """Executes a list of OS commands, and log with logfun.
    
    :param list cmds: list of OS commands (=list of list of str) or a single command (list of str)
    :parma callable logfun: a function to write output, typically 
        ``logging.getLogger('et_micc').debug``.
    :returns int: return code of first failing command, or 0 if all
        commanbds succeed.
    """
    if isinstance(cmds[0],str):
        # this is a single command
        cmds = [cmds]
        
    for cmd in cmds:
        with et_micc_tools.logging_tools.log(logfun, f"> {' '.join(cmd)}"):
            completed_process = subprocess.run(cmd, capture_output=True,env=env)
            if not logfun is None:
                if completed_process.returncode:
                    logfun0 = logfun
                    logfun = logfun.__self__.warning
                    logfun(f"> {' '.join(cmd)}")
                if completed_process.stdout:
                    logfun(' (stdout)\n' + completed_process.stdout.decode('utf-8'))
                if completed_process.stderr:
                    logfun(' (stderr)\n' + completed_process.stderr.decode('utf-8'))
                if completed_process.returncode:
                    logfun = logfun0
            if completed_process.returncode:
                if stop_on_error:
                    return completed_process.returncode
    return 0


def get_project_path(p):
    """Look for a project directory in the parents of path :py:obj:`p`.
    
    :param Path p:
    :returns: the nearest directory above :py:obj:`p` that is project directory.
    :raise: RuntimeError if :py:obj:`p` is noe inside a project directory.
    """
    root = Path('/')
    p = Path(p).resolve()
    p0 = copy.copy(p)
    while not is_project_directory(p):
        p = p.parent
        if p==root:
            raise RuntimeError(f"Folder {p0} is not inside a Python project.")
    return p


def is_conda_python():
    """Test if the user's environment is a conda Python environment.
    
    see https://stackoverflow.com/questions/21282363/any-way-to-tell-if-users-python-environment-is-anaconda
    """
    is_conda = os.path.exists(os.path.join(sys.prefix, 'conda-meta'))
    return is_conda


def is_poetry_available(system):
    """Test if *poetry* is available in the environment."""    
    myenv=os.environ.copy()
    if system:
        cmd=['which','poetry_']
    else:
        cmd=['which','poetry']
    result = subprocess.run(cmd,capture_output=True,env=myenv)
    return result.returncode==0


def is_bumpversion_available():
    """Test if *bumpversion* is available in the environment."""
    myenv=os.environ.copy()
    cmd=['which','bumpversion']
    result = subprocess.run(cmd,capture_output=True,env=myenv)
    return result.returncode==0


def get_dependencies():
    """
    """
    pyproject_toml = TomlFile('pyproject.toml')
    current_deps = pyproject_toml['tool']['poetry']['dependencies']
    return current_deps

# end of file